import { GET, POST } from "../config/route";
import { NotFoundError } from "@snip/database/errors";
import { resetMockConfig, setSessionType } from "__mocks__/mockConfig";
import { createMockRequest } from "__mocks__/request";

const config = {
    userConfigNotfound: false,
};

jest.mock("@snip/database", () => {
    return {
        userConfig: {
            getByUserId: jest.fn(async (id) => {
                if (config.userConfigNotfound) {
                    throw new NotFoundError("User config not found", id);
                }
                return {
                    id: 42,
                    user_id: 42,
                    zoom1: 1,
                    zoom2: 1,
                    text_lineWrap: 1,
                    text_lineHeight: 1,
                    text_fontSize: 1,
                    text_fontColor: "#000000",
                    text_font: "Arial",
                    pen_size: 1,
                    pen_color: "#000000",
                    pen_colors: ["#000000"],
                    pen_smoothing: 0.5,
                };
            }),
            updateByUserId: jest.fn(async (a) => {
                return a;
            }),
        },
    };
});

jest.mock("@snip/database/sql.connection", () => {
    return {
        pool_permissions: {
            query: jest.fn(async () => {
                return [
                    {
                        id: 1,
                    },
                ];
            }),
        },
    };
});

describe("GET api/account/config", () => {
    beforeEach(() => {
        resetMockConfig();
    });

    it("should return 401 if user is not logged in", async () => {
        setSessionType("none");

        const req = await createMockRequest({
            method: "GET",
        });
        const res = await GET(req, {});
        expect(res.status).toBe(401);
    });

    it("should return 200", async () => {
        setSessionType("user");

        const req = await createMockRequest({
            method: "GET",
        });
        const res = await GET(req, {});
        expect(res.status).toBe(200);
        const data = await res.json();
        expect(data).toEqual({
            id: 42,
            user_id: 42,
            zoom1: 1,
            zoom2: 1,
            text_lineWrap: 1,
            text_lineHeight: 1,
            text_fontSize: 1,
            text_fontColor: "#000000",
            text_font: "Arial",
            pen_size: 1,
            pen_color: "#000000",
            pen_colors: ["#000000"],
            pen_smoothing: 0.5,
        });
    });

    it("should return 404 if user config not found", async () => {
        setSessionType("user");
        config.userConfigNotfound = true;

        const req = await createMockRequest({
            method: "GET",
        });
        const res = await GET(req, {});
        expect(res.status).toBe(500);
    });
});

describe("POST api/account/config", () => {
    beforeEach(() => {
        resetMockConfig();
    });

    it("should return 401 if user is not logged in", async () => {
        setSessionType("none");

        const req = await createMockRequest({
            method: "POST",
        });
        const res = await POST(req, {});
        expect(res.status).toBe(401);
    });

    it("should return 200", async () => {
        setSessionType("user");

        const req = await createMockRequest({
            method: "POST",
            body: {
                zoom1: 2,
                zoom2: 2,
                text_lineWrap: 2,
                text_lineHeight: 2,
                text_fontSize: 2,
                text_fontColor: "#111111",
                text_font: "Arial",
                pen_size: 2,
                pen_color: "#111111",
                pen_colors: ["#111111"],
                pen_smoothing: 0.6,
            },
        });
        const res = await POST(req, {});
        expect(res.status).toBe(200);
    });

    it("should return 400 if no updates given", async () => {
        setSessionType("user");

        const req = await createMockRequest({
            method: "POST",
            body: {},
        });
        const res = await POST(req, {});
        expect(res.status).toBe(400);
    });

    it("should allow for partial updates", async () => {
        setSessionType("user");

        const values = {
            zoom1: 2,
            zoom2: 1,
            text_lineWrap: 1,
            text_lineHeight: 1,
            text_fontSize: 1,
            text_fontColor: "#000000",
            text_font: "Arial",
            pen_size: 1,
            pen_color: "#000000",
            pen_colors: ["#000000"],
            pen_smoothing: 0.5,
        };

        for (const key in values) {
            const req = await createMockRequest({
                method: "POST",
                body: {
                    [key]: values[key],
                },
            });

            const res = await POST(req, {});
            expect(res.status).toBe(200);
        }
    });

    it("should return 400 if request is invalid", async () => {
        setSessionType("user");

        const req = await createMockRequest({
            method: "POST",
            body: {
                zoom1: "a",
            },
        });
        const res = await POST(req, {});
        expect(res.status).toBe(400);
    });

    it("should return 400 if bad font", async () => {
        setSessionType("user");

        const req = await createMockRequest({
            method: "POST",
            body: {
                text_font: "a",
            },
        });
        const res = await POST(req, {});
        expect(res.status).toBe(400);
    });
});
