import { IncomingMessage, ServerResponse } from "http";
import { getIronSession, IronSession, sealData } from "iron-session";

import { config } from "@snip/config/server";

// We use the same session data in the frontend and backend
// this makes it a bit easier to keep track of who is
// authenticated and what they are allowed to do

interface SessionData {
    user?: UserSessionData;
    anonymous?: TokenSessionData;
    api?: ApiSessionData;
}

/** For user sessions
 * they create a session for a user
 * e.g. login
 */
interface UserSessionData {
    id: number;
    email_verified: boolean;
    created: number;
}

/** For internal request
 * they create a session for an api user
 * e.g. pdf render in parallel
 */
interface ApiSessionData {
    // Identifier for the api user
    identifier: string;
}

/** For ui tokens
 * they also create a session
 * for the (anonymous) user
 * e.g. sharable links
 * we allow multiple ui tokens
 * to be active at the same time
 */
export interface TokenSessionData {
    tokens: {
        id: number;
        book_id: number;
        expires_at: number;
    }[];
}

type Prettify<T> = {
    [K in keyof T]: T[K];
} & object;

export type SnipSessionData = Prettify<IronSession<SessionData>>;

export const sessionOptions = {
    password: config.SESSION_SECRET as string,
    cookieName: "snip_labbook",
    // secure: true should be used in production (HTTPS) but can't be used in development (HTTP)
    cookieOptions: {
        secure: true,
        domain: config.HOSTNAME,
    },
};

/** Get session in the route handlers
 * @param req
 * @param res
 */
export async function getSession(
    req: IncomingMessage | Request,
    res: Response | ServerResponse,
): Promise<SnipSessionData>;
export async function getSession(cookies: unknown): Promise<SnipSessionData>;

// Combined implementation of getSession
export async function getSession(
    arg1: IncomingMessage | Request,
    arg2?: Response | ServerResponse,
): Promise<SnipSessionData> {
    if (arg2) {
        // If the second argument is present, assume the first one is Request and the second one is Response
        const req = arg1 as IncomingMessage;
        const res = arg2 as Response;
        const session = await getIronSession<SnipSessionData>(
            req,
            res,
            sessionOptions,
        );
        return session;
    } else {
        // If the second argument is not present, assume the first one is cookies
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const cookies = arg1 as any;
        const session = await getIronSession<SnipSessionData>(
            cookies,
            sessionOptions,
        );
        return session;
    }
}

/** Get the session string for an internal api request
 * @param data
 *
 * For example useful for an internal api request
 * ```typescript
 * const session = await getCookieString({ user_api: "socket" });
 * fetch("http://localhost:4000/api/v1/snip", {
 *    headers: {
 *       "Content-Type": "application/json",
 *       "cookie": `snip_labbook=${session}`
 *   }
 * });
 */
export async function getCookieString(data: SessionData): Promise<string> {
    return await sealData(data, sessionOptions).then((data) => {
        return `${sessionOptions.cookieName}=${data}`;
    });
}
